@@ -1,7 +1,6 @@
 #include <Python.h>
 #include <structmember.h>
-#include <kdumpfile.h>
-#include <addrxlat.h>
+#include <libkdumpfile/kdumpfile.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <sys/stat.h>
@@ -44,6 +43,10 @@ static PyTypeObject attr_iterkey_object_type;
 static PyTypeObject attr_itervalue_object_type;
 static PyTypeObject attr_iteritem_object_type;
 
+static PyTypeObject bmp_object_type;
+
+static PyTypeObject blob_object_type;
+
 static PyObject *attr_viewkeys_type;
 static PyObject *attr_viewvalues_type;
 static PyObject *attr_viewitems_type;
@@ -52,6 +55,10 @@ static PyObject *attr_viewdict_type;
 static PyObject *attr_dir_new(kdumpfile_object *kdumpfile,
 			      const kdump_attr_ref_t *baseref);
 
+static PyObject *bmp_new(kdump_bmp_t *bitmap);
+
+static PyObject *blob_new(kdump_blob_t *blob);
+
 static PyObject *
 exception_map(kdump_status status)
 {
@@ -182,7 +189,8 @@ static PyObject *kdumpfile_read (PyObject *_self, PyObject *args, PyObject *kw)
 			    PyByteArray_AS_STRING(obj), &r);
 	if (status != KDUMP_OK) {
 		Py_XDECREF(obj);
-		PyErr_Format(exception_map(status), kdump_get_err(self->ctx));
+		PyErr_SetString(exception_map(status),
+				kdump_get_err(self->ctx));
 		return NULL;
 	}
 
@@ -204,6 +212,10 @@ attr_new(kdumpfile_object *kdumpfile, kdump_attr_ref_t *ref, kdump_attr_t *attr)
 			return PyString_FromString(attr->val.string);
 		case KDUMP_DIRECTORY:
 			return attr_dir_new(kdumpfile, ref);
+		case KDUMP_BITMAP:
+			return bmp_new(attr->val.bitmap);
+		case KDUMP_BLOB:
+			return blob_new(attr->val.blob);
 		default:
 			PyErr_SetString(PyExc_RuntimeError, "Unhandled attr type");
 			return NULL;
@@ -217,7 +229,15 @@ static PyObject *
 get_addrxlat_ctx(PyObject *_self, PyObject *args)
 {
 	kdumpfile_object *self = (kdumpfile_object*)_self;
-	addrxlat_ctx_t *ctx = kdump_get_addrxlat_ctx(self->ctx);
+	addrxlat_ctx_t *ctx;
+	kdump_status status;
+
+	status = kdump_get_addrxlat(self->ctx, &ctx, NULL);
+	if (status != KDUMP_OK) {
+		PyErr_SetString(exception_map(status),
+				kdump_get_err(self->ctx));
+		return NULL;
+	}
 	return addrxlat_API->Context_FromPointer(self->addrxlat_convert, ctx);
 }
 
@@ -228,7 +248,15 @@ static PyObject *
 get_addrxlat_sys(PyObject *_self, PyObject *args)
 {
 	kdumpfile_object *self = (kdumpfile_object*)_self;
-	addrxlat_sys_t *sys = kdump_get_addrxlat_sys(self->ctx);
+	addrxlat_sys_t *sys;
+	kdump_status status;
+
+	status = kdump_get_addrxlat(self->ctx, NULL, &sys);
+	if (status != KDUMP_OK) {
+		PyErr_SetString(exception_map(status),
+				kdump_get_err(self->ctx));
+		return NULL;
+	}
 	return addrxlat_API->System_FromPointer(self->addrxlat_convert, sys);
 }
 
@@ -521,7 +549,7 @@ attr_dir_length(PyObject *_self)
 	return len;
 
  err:
-	PyErr_Format(exception_map(status), kdump_get_err(ctx));
+	PyErr_SetString(exception_map(status), kdump_get_err(ctx));
 	return -1;
 }
 
@@ -1019,7 +1047,7 @@ attr_dir_clear(PyObject *_self, PyObject *args)
  err:
 	kdump_attr_iter_end(ctx, &iter);
  err_noiter:
-	PyErr_Format(exception_map(status), kdump_get_err(ctx));
+	PyErr_SetString(exception_map(status), kdump_get_err(ctx));
 	return NULL;
 }
 
@@ -1038,7 +1066,7 @@ attr_dir_repr(PyObject *_self)
 
 	status = kdump_attr_ref_iter_start(ctx, &self->baseref, &iter);
 	if (status != KDUMP_OK) {
-		PyErr_Format(exception_map(status), kdump_get_err(ctx));
+		PyErr_SetString(exception_map(status), kdump_get_err(ctx));
 		return NULL;
 	}
 
@@ -1080,7 +1108,8 @@ attr_dir_repr(PyObject *_self)
 
 		status = kdump_attr_iter_next(ctx, &iter);
 		if (status != KDUMP_OK) {
-			PyErr_Format(exception_map(status), kdump_get_err(ctx));
+			PyErr_SetString(exception_map(status),
+					kdump_get_err(ctx));
 			goto out;
 		}
 	}
@@ -1129,7 +1158,7 @@ attr_dir_print(PyObject *_self, FILE *fp, int flags)
 
 	status = kdump_attr_ref_iter_start(ctx, &self->baseref, &iter);
 	if (status != KDUMP_OK) {
-		PyErr_Format(exception_map(status), kdump_get_err(ctx));
+		PyErr_SetString(exception_map(status), kdump_get_err(ctx));
 		return -1;
 	}
 
@@ -1162,7 +1191,8 @@ attr_dir_print(PyObject *_self, FILE *fp, int flags)
 
 		status = kdump_attr_iter_next(ctx, &iter);
 		if (status != KDUMP_OK) {
-			PyErr_Format(exception_map(status), kdump_get_err(ctx));
+			PyErr_SetString(exception_map(status),
+					kdump_get_err(ctx));
 			goto err;
 		}
 
@@ -1468,7 +1498,7 @@ attr_iter_new(attr_dir_object *attr_dir, PyTypeObject *itertype)
 	status = kdump_attr_ref_iter_start(ctx, &attr_dir->baseref,
 					   &self->iter);
 	if (status != KDUMP_OK) {
-		PyErr_Format(exception_map(status), kdump_get_err(ctx));
+		PyErr_SetString(exception_map(status), kdump_get_err(ctx));
 		Py_DECREF(self);
 		return NULL;
 	}
@@ -1508,7 +1538,7 @@ attr_iter_advance(attr_iter_object *self, PyObject *ret)
 
 	status = kdump_attr_iter_next(ctx, &self->iter);
 	if (status != KDUMP_OK) {
-		PyErr_Format(exception_map(status), kdump_get_err(ctx));
+		PyErr_SetString(exception_map(status), kdump_get_err(ctx));
 		Py_XDECREF(ret);
 		ret = NULL;
 	}
@@ -1683,6 +1713,348 @@ static PyTypeObject attr_iteritem_object_type = {
 	attr_iteritem_next,		/* tp_iternext */
 };
 
+typedef struct {
+	PyObject_HEAD
+	kdump_bmp_t *bmp;
+} bmp_object;
+
+PyDoc_STRVAR(bmp__doc__,
+"bmp() -> dump bitmap");
+
+static PyObject *
+bmp_new(kdump_bmp_t *bmp)
+{
+	bmp_object *self;
+
+	self = PyObject_New(bmp_object, &bmp_object_type);
+	if (!self)
+		return NULL;
+
+	kdump_bmp_incref(bmp);
+	self->bmp = bmp;
+
+	return (PyObject*)self;
+}
+
+static void
+bmp_dealloc(PyObject *_self)
+{
+	bmp_object *self = (bmp_object*)_self;
+
+	if (self->bmp)
+		kdump_bmp_decref(self->bmp);
+
+	Py_TYPE(self)->tp_free((PyObject*)self);
+}
+
+PyDoc_STRVAR(bmp_get_bits__doc__,
+"BMP.get_bits(first, last) -> byte array\n\
+\n\
+Get bitmap bits as a raw bitmap.");
+
+static PyObject *
+bmp_get_bits(PyObject *_self, PyObject *args, PyObject *kwargs)
+{
+	static char *keywords[] = {"first", "last", NULL};
+	bmp_object *self = (bmp_object*)_self;
+	unsigned long long first, last;
+	PyObject *buffer;
+	Py_ssize_t sz;
+	kdump_status status;
+
+	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "KK:get_bits",
+					 keywords, &first, &last))
+		return NULL;
+
+	buffer = PyByteArray_FromStringAndSize(NULL, 0);
+	if (!buffer)
+		return NULL;
+
+	sz = (((last - first) | 7) + 1) / 8;
+	if (PyByteArray_Resize(buffer, sz) < 0) {
+		Py_DECREF(buffer);
+		return NULL;
+	}
+
+
+	status = kdump_bmp_get_bits(
+		self->bmp, first, last,
+		(unsigned char*)PyByteArray_AS_STRING(buffer));
+	if (status != KDUMP_OK) {
+		Py_DECREF(buffer);
+		PyErr_SetString(exception_map(status),
+				kdump_bmp_get_err(self->bmp));
+		return NULL;
+	}
+
+	return buffer;
+}
+
+PyDoc_STRVAR(bmp_find_set__doc__,
+"BMP.find_set(idx) -> index\n\
+\n\
+Find the closest set bit in a bitmapm, starting at idx.");
+
+static PyObject *
+bmp_find_set(PyObject *_self, PyObject *args, PyObject *kwargs)
+{
+	static char *keywords[] = {"idx", NULL};
+	bmp_object *self = (bmp_object*)_self;
+	unsigned long long argidx;
+	kdump_addr_t idx;
+	kdump_status status;
+
+	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "K:find_set",
+					 keywords, &argidx))
+		return NULL;
+
+	idx = argidx;
+	status = kdump_bmp_find_set(self->bmp, &idx);
+	if (status != KDUMP_OK) {
+		PyErr_SetString(exception_map(status),
+				kdump_bmp_get_err(self->bmp));
+		return NULL;
+	}
+
+	return PyLong_FromUnsignedLong(idx);
+}
+
+PyDoc_STRVAR(bmp_find_clear__doc__,
+"BMP.find_clear(idx) -> index\n\
+\n\
+Find the closest zero bit in a bitmapm, starting at idx.");
+
+static PyObject *
+bmp_find_clear(PyObject *_self, PyObject *args, PyObject *kwargs)
+{
+	static char *keywords[] = {"idx", NULL};
+	bmp_object *self = (bmp_object*)_self;
+	unsigned long long argidx;
+	kdump_addr_t idx;
+	kdump_status status;
+
+	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "K:find_clear",
+					 keywords, &argidx))
+		return NULL;
+
+	idx = argidx;
+	status = kdump_bmp_find_clear(self->bmp, &idx);
+	if (status != KDUMP_OK) {
+		PyErr_SetString(exception_map(status),
+				kdump_bmp_get_err(self->bmp));
+		return NULL;
+	}
+
+	return PyLong_FromUnsignedLong(idx);
+}
+
+static PyMethodDef bmp_methods[] = {
+	{ "get_bits", (PyCFunction)bmp_get_bits,
+	  METH_VARARGS | METH_KEYWORDS,
+	  bmp_get_bits__doc__ },
+	{ "find_set", (PyCFunction)bmp_find_set,
+	  METH_VARARGS | METH_KEYWORDS,
+	  bmp_find_set__doc__ },
+	{ "find_clear", (PyCFunction)bmp_find_clear,
+	  METH_VARARGS | METH_KEYWORDS,
+	  bmp_find_clear__doc__ },
+	{NULL,		NULL}	/* sentinel */
+};
+
+static PyTypeObject bmp_object_type =
+{
+	PyVarObject_HEAD_INIT(NULL, 0)
+	MOD_NAME ".bmp",		/* tp_name */
+	sizeof (bmp_object),		/* tp_basicsize */
+	0,				/* tp_itemsize */
+	bmp_dealloc,			/* tp_dealloc */
+	0,				/* tp_print */
+	0,				/* tp_getattr */
+	0,				/* tp_setattr */
+	0,				/* tp_compare */
+	0,				/* tp_repr */
+	0,				/* tp_as_number */
+	0,				/* tp_as_sequence */
+	0,				/* tp_as_mapping */
+	0,				/* tp_hash */
+	0,				/* tp_call */
+	0,				/* tp_str */
+	0,				/* tp_getattro */
+	0,				/* tp_setattro */
+	0,				/* tp_as_buffer */
+	Py_TPFLAGS_DEFAULT
+	    | Py_TPFLAGS_BASETYPE,	/* tp_flags */
+	bmp__doc__,			/* tp_doc */
+	0,				/* tp_traverse */
+	0,				/* tp_clear */
+	0,				/* tp_richcompare */
+	0,				/* tp_weaklistoffset */
+	0,				/* tp_iter */
+	0,				/* tp_iternext */
+	bmp_methods,			/* tp_methods */
+};
+
+typedef struct {
+	PyObject_HEAD
+	kdump_blob_t *blob;
+} blob_object;
+
+PyDoc_STRVAR(blob__doc__,
+"blob() -> dump blob");
+
+static PyObject *
+blob_new(kdump_blob_t *blob)
+{
+	blob_object *self;
+
+	self = PyObject_New(blob_object, &blob_object_type);
+	if (!self)
+		return NULL;
+
+	kdump_blob_incref(blob);
+	self->blob = blob;
+
+	return (PyObject*)self;
+}
+
+static void
+blob_dealloc(PyObject *_self)
+{
+	blob_object *self = (blob_object*)_self;
+
+	if (self->blob)
+		kdump_blob_decref(self->blob);
+
+	Py_TYPE(self)->tp_free((PyObject*)self);
+}
+
+PyDoc_STRVAR(blob_set__doc__,
+"BLOB.set(buffer)\n\
+\n\
+Replace blob contents with a new value. The object given as\n\
+argument must implement the buffer protocol.");
+
+static PyObject *
+blob_set(PyObject *_self, PyObject *args)
+{
+	blob_object *self = (blob_object*)_self;
+	kdump_status status;
+	Py_buffer view;
+	PyObject *arg;
+	void *buffer;
+
+	if (!PyArg_ParseTuple(args, "O:set", &arg))
+		return NULL;
+
+	if (!PyObject_CheckBuffer(arg)) {
+		PyErr_Format(PyExc_TypeError,
+			     "Type %.100s doesn't support the buffer API",
+			     Py_TYPE(arg)->tp_name);
+		return NULL;
+	}
+
+	if (PyObject_GetBuffer(arg, &view, PyBUF_FULL_RO) < 0)
+		return NULL;
+	buffer = malloc(view.len);
+	if (!buffer)
+		goto buf_fail;
+	if (PyBuffer_ToContiguous(buffer, &view, view.len, 'C') < 0)
+		goto buf_fail;
+	PyBuffer_Release(&view);
+
+	status = kdump_blob_set(self->blob, buffer, view.len);
+	if (status != KDUMP_OK) {
+		PyErr_SetString(exception_map(status),
+				kdump_strerror(status));
+		free(buffer);
+		return NULL;
+	}
+	return Py_None;
+
+ buf_fail:
+	PyBuffer_Release(&view);
+	return NULL;
+}
+
+static PyMethodDef blob_methods[] = {
+	{ "set", blob_set, METH_VARARGS,
+	  blob_set__doc__ },
+	{NULL,		NULL}	/* sentinel */
+};
+
+static int
+blob_getbuffer(PyObject *_self, Py_buffer *view, int flags)
+{
+	blob_object *self = (blob_object*)_self;
+	void *buffer;
+	size_t size;
+	int ret;
+
+	buffer = kdump_blob_pin(self->blob);
+	if (view == NULL)
+		return 0;
+
+	size = kdump_blob_size(self->blob);
+	ret = PyBuffer_FillInfo(view, _self, buffer, size, 0, flags);
+	if (ret < 0)
+		kdump_blob_unpin(self->blob);
+	return ret;
+}
+
+static void
+blob_releasebuffer(PyObject *_self, Py_buffer *view)
+{
+	blob_object *self = (blob_object*)_self;
+	kdump_blob_unpin(self->blob);
+}
+
+static PyBufferProcs blob_as_buffer = {
+#if PY_MAJOR_VERSION < 3
+    (readbufferproc)NULL,
+    (writebufferproc)NULL,
+    (segcountproc)NULL,
+    (charbufferproc)NULL,
+#endif
+    blob_getbuffer,
+    blob_releasebuffer,
+};
+
+static PyTypeObject blob_object_type =
+{
+	PyVarObject_HEAD_INIT(NULL, 0)
+	MOD_NAME ".blob",		/* tp_name */
+	sizeof (blob_object),		/* tp_basicsize */
+	0,				/* tp_itemsize */
+	blob_dealloc,			/* tp_dealloc */
+	0,				/* tp_print */
+	0,				/* tp_getattr */
+	0,				/* tp_setattr */
+	0,				/* tp_compare */
+	0,				/* tp_repr */
+	0,				/* tp_as_number */
+	0,				/* tp_as_sequence */
+	0,				/* tp_as_mapping */
+	0,				/* tp_hash */
+	0,				/* tp_call */
+	0,				/* tp_str */
+	0,				/* tp_getattro */
+	0,				/* tp_setattro */
+	&blob_as_buffer,		/* tp_as_buffer */
+	Py_TPFLAGS_DEFAULT
+#if PY_MAJOR_VERSION < 3
+	    | Py_TPFLAGS_HAVE_NEWBUFFER
+#endif
+	    | Py_TPFLAGS_BASETYPE,	/* tp_flags */
+	blob__doc__,			/* tp_doc */
+	0,				/* tp_traverse */
+	0,				/* tp_clear */
+	0,				/* tp_richcompare */
+	0,				/* tp_weaklistoffset */
+	0,				/* tp_iter */
+	0,				/* tp_iternext */
+	blob_methods,			/* tp_methods */
+};
+
 struct constdef {
 	const char *name;
 	int value;
@@ -1738,6 +2110,10 @@ init_kdumpfile (void)
 		return MOD_ERROR_VAL;
 	if (PyType_Ready(&attr_iteritem_object_type) < 0)
 		return MOD_ERROR_VAL;
+	if (PyType_Ready(&bmp_object_type) < 0)
+		return MOD_ERROR_VAL;
+	if (PyType_Ready(&blob_object_type) < 0)
+		return MOD_ERROR_VAL;
 
 #if PY_MAJOR_VERSION >= 3
 	mod = PyModule_Create(&kdumpfile_moddef);
@@ -1759,6 +2135,18 @@ init_kdumpfile (void)
 	if (ret)
 		goto fail;
 
+	Py_INCREF((PyObject *)&bmp_object_type);
+	ret = PyModule_AddObject(mod, "bmp",
+				 (PyObject*)&bmp_object_type);
+	if (ret)
+		goto fail;
+
+	Py_INCREF((PyObject *)&blob_object_type);
+	ret = PyModule_AddObject(mod, "blob",
+				 (PyObject*)&blob_object_type);
+	if (ret)
+		goto fail;
+
 	for (cdef = kdumpfile_constants; cdef->name; ++cdef)
 		if (PyModule_AddIntConstant(mod, cdef->name, cdef->value))
 			goto fail;
