@@ -3,6 +3,10 @@
 
 import unittest
 import addrxlat
+import sys
+
+if (sys.version_info.major >= 3):
+    xrange = range
 
 class TestFullAddress(unittest.TestCase):
     def test_fulladdr_defaults(self):
@@ -273,11 +277,11 @@ class TestMethod(unittest.TestCase):
             meth.tbl = ((1,),)
         with self.assertRaisesRegexp(ValueError, 'must be integer pairs'):
             meth.tbl = ((1, 2, 3),)
-        with self.assertRaisesRegexp(TypeError, 'not an integer'):
+        with self.assertRaisesRegexp(TypeError, 'must be.* a number'):
             meth.tbl = ((None, None),)
-        with self.assertRaisesRegexp(TypeError, 'not an integer'):
+        with self.assertRaisesRegexp(TypeError, 'must be.* a number'):
             meth.tbl = ((1, None),)
-        with self.assertRaisesRegexp(TypeError, 'not an integer'):
+        with self.assertRaisesRegexp(TypeError, 'must be.* a number'):
             meth.tbl = ((None, 1),)
 
     def test_memarr_defaults(self):
@@ -575,7 +579,7 @@ class TestStep(unittest.TestCase):
             step.idx = None
         with self.assertRaisesRegexp(ValueError, 'more than [0-9]+ indices'):
             step.idx = (0,) * (addrxlat.FIELDS_MAX + 2)
-        with self.assertRaisesRegexp(TypeError, 'not an integer'):
+        with self.assertRaisesRegexp(TypeError, 'must be.* a number'):
             step.idx = (None,)
 
 class TestOperator(unittest.TestCase):
@@ -631,23 +635,27 @@ class TestOperator(unittest.TestCase):
 #
 class TestTranslation(unittest.TestCase):
     def setUp(self):
-        def read32(addr):
+        def get_page(addr):
             # Page table level 2 @ 0
             if addr.addr == 0x10000:
-                return 0x101
+                return (bytearray((0x00, 0x00, 0x01, 0x01)),
+                        addrxlat.BIG_ENDIAN)
             # Page table level 1 @ 0x65
             if addr.addr == 0x10100 + 0x65 * 4:
-                return 0x1c0
+                return (bytearray((0x00, 0x00, 0x01, 0xc0)),
+                        addrxlat.BIG_ENDIAN)
             # Page table level 1 @ 0x41
             if addr.addr == 0x10100 + 0x41 * 4:
-                return 0x1a9
+                return (bytearray((0x00, 0x00, 0x01, 0xa9)),
+                        addrxlat.BIG_ENDIAN)
             # Memory array at 0x40
             if addr.addr == 0x11000 + 0x40 * 4:
-                return 0xaa
+                return (bytearray((0x00, 0x00, 0x00, 0xaa)),
+                        addrxlat.BIG_ENDIAN)
 
         self.ctx = addrxlat.Context()
         self.ctx.read_caps = addrxlat.CAPS(addrxlat.MACHPHYSADDR)
-        self.ctx.cb_read32 = read32
+        self.ctx.cb_get_page = get_page
         self.sys = addrxlat.System()
 
         map = addrxlat.Map()
@@ -783,10 +791,11 @@ class TestTranslation(unittest.TestCase):
             def __init__(self, *args, **kwargs):
                 super(mycontext, self).__init__(*args, **kwargs)
                 self.read_caps = addrxlat.CAPS(addrxlat.MACHPHYSADDR)
-            def cb_read32(self, addr):
+            def cb_get_page(self, addr):
                 # Memory array at 0x40
                 if addr.addr == 0x11000 + 0x40 * 4:
-                    return 0x12
+                    return (bytearray((0x00, 0x00, 0x00, 0x12)),
+                            addrxlat.BIG_ENDIAN)
 
         ctx = mycontext()
         addr = addrxlat.FullAddress(addrxlat.KVADDR, 0x4034)
