@@ -28,6 +28,8 @@
    not, see <http://www.gnu.org/licenses/>.
 */
 
+#include <string.h>
+
 #include "addrxlat-priv.h"
 
 /** Count total size of all address bitfields.
@@ -60,44 +62,6 @@ paging_max_index(const addrxlat_paging_form_t *pf)
 		: ADDRXLAT_ADDR_MAX);
 }
 
-/** Read the raw PTE value.
- * @param step  Current step state.
- * @returns     Error status.
- *
- * On successful return, @c step->raw.pte contains the raw
- * PTE value for the current translation step.
- */
-addrxlat_status
-read_pte(addrxlat_step_t *step)
-{
-	uint64_t pte64;
-	uint32_t pte32;
-	addrxlat_pte_t pte;
-	addrxlat_status status;
-
-	switch(step->elemsz) {
-	case 4:
-		status = read32(step, &step->base, &pte32, "PTE");
-		pte = pte32;
-		break;
-
-	case 8:
-		status = read64(step, &step->base, &pte64, "PTE");
-		pte = pte64;
-		break;
-
-	default:
-		return set_error(step->ctx, ADDRXLAT_ERR_NOTIMPL,
-				 "Unsupported PTE size: %u",
-				 step->elemsz);
-	}
-
-	if (status == ADDRXLAT_OK)
-		step->raw.pte = pte;
-
-	return status;
-}
-
 /** Update current step state for huge page.
  * @param step  Current step state.
  * @returns     Always @c ADDRXLAT_OK.
@@ -173,28 +137,60 @@ first_step_pgt_generic(addrxlat_step_t *step, addrxlat_addr_t addr)
 	return ADDRXLAT_OK;
 }
 
-/** Page frame number next step function.
+/** Common next step helper for page frame number.
+ * @param step  Current step state.
+ * @param pte   Page table entry value (possibly masked).
+ * @returns     Error status.
+ *
+ * This function contains the common processing for 32-bit and
+ * 64-bit PFNs after the PTE was read.
+ */
+static addrxlat_status
+next_step_pfn_common(addrxlat_step_t *step, addrxlat_pte_t pte)
+{
+	const addrxlat_meth_t *meth;
+	if (!pte)
+		return set_error(step->ctx, ADDRXLAT_ERR_NOTPRESENT,
+				 "Level-%u PFN not present",
+				 step->remain);
+	meth = step->meth;
+	step->base.addr = pte << meth->param.pgt.pf.fieldsz[0];
+	step->base.as = meth->target_as;
+	if (step->remain == 1)
+		step->elemsz = 1;
+	return ADDRXLAT_OK;
+}
+
+/** Next step function for a 32-bit page frame number.
  * @param step  Current step state.
  * @returns     Error status.
  *
- * This function handles page frame numbers in a table.
+ * This function handles 32-bit page frame numbers in a table.
  */
 static addrxlat_status
-next_step_pfn(addrxlat_step_t *step)
+next_step_pfn32(addrxlat_step_t *step)
 {
-	addrxlat_status status;
+	addrxlat_pte_t pte;
+	addrxlat_status status = read_pte32(step, &pte);
+	if (status != ADDRXLAT_OK)
+		return status;
+	return next_step_pfn_common(step, pte);
+}
 
-	status = read_pte(step);
-	if (status == ADDRXLAT_OK) {
-		const addrxlat_meth_t *meth = step->meth;
-		step->base.addr =
-			step->raw.pte << meth->param.pgt.pf.fieldsz[0];
-		step->base.as = meth->target_as;
-		if (step->remain == 1)
-			step->elemsz = 1;
-	}
-
-	return status;
+/** Next step function for a 64-bit page frame number.
+ * @param step  Current step state.
+ * @returns     Error status.
+ *
+ * This function handles 64-bit page frame numbers in a table.
+ */
+static addrxlat_status
+next_step_pfn64(addrxlat_step_t *step)
+{
+	addrxlat_pte_t pte;
+	addrxlat_status status = read_pte64(step, &pte);
+	if (status != ADDRXLAT_OK)
+		return status;
+	return next_step_pfn_common(step, pte);
 }
 
 /** Check unsigned address overflow.
@@ -280,6 +276,7 @@ first_step_pgt(addrxlat_step_t *step, addrxlat_addr_t addr)
 {
 	switch (step->meth->param.pgt.pf.pte_format) {
 	case ADDRXLAT_PTE_NONE:
+	case ADDRXLAT_PTE_AARCH64:
 	case ADDRXLAT_PTE_PPC64_LINUX_RPN30:
 		return first_step_pgt_generic(step, addr);
 
@@ -312,8 +309,13 @@ next_step_pgt(addrxlat_step_t *step)
 		return ADDRXLAT_OK;
 
 	case ADDRXLAT_PTE_PFN32:
+		return next_step_pfn32(step);
+
 	case ADDRXLAT_PTE_PFN64:
-		return next_step_pfn(step);
+		return next_step_pfn64(step);
+
+	case ADDRXLAT_PTE_AARCH64:
+		return pgt_aarch64(step);
 
 	case ADDRXLAT_PTE_IA32:
 		return pgt_ia32(step);
@@ -524,11 +526,292 @@ addrxlat_walk(addrxlat_step_t *step)
 
 	clear_error(step->ctx);
 
-	while (step->remain) {
-		status = internal_step(step);
+	status = first_step(step, step->base.addr);
+	if (status != ADDRXLAT_OK)
+		return status;
+
+	while (--step->remain) {
+		step->base.addr += step->idx[step->remain] * step->elemsz;
+		status = next_step(step);
 		if (status != ADDRXLAT_OK)
-			break;
+			return status;
 	}
 
+	step->base.as = step->meth->target_as;
+	step->base.addr += step->idx[0] * step->elemsz;
+	step->elemsz = 0;
+	return ADDRXLAT_OK;
+}
+
+/** Find the lowest mapped virtual address in a given page table.
+ * @param step   Current step state.
+ * @param addr   First address to try; updated on return.
+ * @param limit  Last address to try.
+ * @returns      Error status.
+ */
+static addrxlat_status
+lowest_mapped_tbl(addrxlat_step_t *step,
+		  addrxlat_addr_t *addr, addrxlat_addr_t limit)
+{
+	int i;
+	addrxlat_addr_t nelem;
+	addrxlat_addr_t tblmask;
+	addrxlat_step_t mystep;
+	addrxlat_status status;
+
+	nelem = pf_table_size(&step->meth->param.pgt.pf, step->remain - 1);
+	tblmask = pf_table_mask(&step->meth->param.pgt.pf, step->remain - 1);
+	memcpy(&mystep, step, sizeof *step);
+	while (*addr <= limit) {
+		status = internal_step(step);
+		if (status == ADDRXLAT_OK) {
+			if (step->remain <= 1)
+				return internal_step(step);
+
+			status = lowest_mapped_tbl(step, addr, limit);
+			if (status != ADDRXLAT_ERR_NOTPRESENT)
+				return status;
+		} else if (status == ADDRXLAT_ERR_NOTPRESENT) {
+			clear_error(step->ctx);
+			*addr = (*addr | tblmask) + 1;
+		} else
+			return status;
+
+		for (i = 0; i < mystep.remain - 1; ++i)
+			mystep.idx[i] = 0;
+		if (++mystep.idx[i] >= nelem) {
+			bury_cache_buffer(&step->ctx->cache, &mystep.base);
+			return ADDRXLAT_ERR_NOTPRESENT;
+		}
+		memcpy(step, &mystep, sizeof *step);
+	}
+
+	return ADDRXLAT_ERR_NOTPRESENT;
+}
+
+/** Find the lowest mapped virtual address in a given range.
+ * @param step   Initial step state.
+ * @param addr   First address to try; updated on return.
+ * @param limit  Last address to try.
+ * @returns      Error status.
+ *
+ * The initial step state must be initialized same way as for a call
+ * to @ref addrxlat_launch.
+ */
+addrxlat_status
+lowest_mapped(addrxlat_step_t *step,
+	      addrxlat_addr_t *addr, addrxlat_addr_t limit)
+{
+	addrxlat_addr_t page_mask;
+	int savednoerr;
+	addrxlat_status status;
+
+	page_mask = pf_page_mask(&step->meth->param.pgt.pf);
+	*addr &= ~page_mask;
+
+	status = internal_launch(step, *addr);
+	if (status != ADDRXLAT_OK)
+		return status;
+
+	savednoerr = step->ctx->noerr.notpresent;
+	step->ctx->noerr.notpresent = 1;
+	status = lowest_mapped_tbl(step, addr, limit);
+	step->ctx->noerr.notpresent = savednoerr;
 	return status;
 }
+
+/** Find the highest mapped virtual address in a given page table.
+ * @param step   Current step state.
+ * @param addr   First address to try; updated on return.
+ * @param limit  Last address to try.
+ * @returns      Error status.
+ */
+static addrxlat_status
+highest_mapped_tbl(addrxlat_step_t *step,
+		   addrxlat_addr_t *addr, addrxlat_addr_t limit)
+{
+	int i;
+	addrxlat_addr_t nelem;
+	addrxlat_addr_t tblmask;
+	addrxlat_step_t mystep;
+	addrxlat_status status;
+
+	nelem = pf_table_size(&step->meth->param.pgt.pf, step->remain - 1);
+	tblmask = pf_table_mask(&step->meth->param.pgt.pf, step->remain - 1);
+	memcpy(&mystep, step, sizeof *step);
+	while (*addr >= limit) {
+		status = internal_step(step);
+		if (status == ADDRXLAT_OK) {
+			if (step->remain <= 1)
+				return internal_step(step);
+
+			status = highest_mapped_tbl(step, addr, limit);
+			if (status != ADDRXLAT_ERR_NOTPRESENT)
+				return status;
+		} else if (status == ADDRXLAT_ERR_NOTPRESENT) {
+			clear_error(step->ctx);
+			*addr = (*addr & ~tblmask) - 1;
+		} else
+			return status;
+
+		for (i = 0; i < mystep.remain - 1; ++i)
+			mystep.idx[i] = nelem - 1;
+		if (!mystep.idx[i]--) {
+			bury_cache_buffer(&step->ctx->cache, &mystep.base);
+			return ADDRXLAT_ERR_NOTPRESENT;
+		}
+		memcpy(step, &mystep, sizeof *step);
+	}
+
+	return ADDRXLAT_ERR_NOTPRESENT;
+}
+
+/** Find the highest mapped virtual address in a given range.
+ * @param step   Initial step state.
+ * @param addr   First address to try; updated on return.
+ * @param limit  Last address to try.
+ * @returns      Error status.
+ *
+ * The initial step state must be initialized same way as for a call
+ * to @ref addrxlat_launch.
+ */
+addrxlat_status
+highest_mapped(addrxlat_step_t *step,
+	       addrxlat_addr_t *addr, addrxlat_addr_t limit)
+{
+	addrxlat_addr_t page_mask;
+	int savednoerr;
+	addrxlat_status status;
+
+	page_mask = pf_page_mask(&step->meth->param.pgt.pf);
+	*addr |= page_mask;
+
+	status = internal_launch(step, *addr);
+	if (status != ADDRXLAT_OK)
+		return status;
+
+	savednoerr = step->ctx->noerr.notpresent;
+	step->ctx->noerr.notpresent = 1;
+	status = highest_mapped_tbl(step, addr, limit);
+	step->ctx->noerr.notpresent = savednoerr;
+	return status;
+}
+
+/** Find the lowest unmapped virtual address in a given page table.
+ * @param step   Current step state.
+ * @param addr   First address to try; updated on return.
+ * @param limit  Last address to try.
+ * @returns      Error status.
+ */
+static addrxlat_status
+lowest_unmapped_tbl(addrxlat_step_t *step,
+		    addrxlat_addr_t *addr, addrxlat_addr_t limit)
+{
+	int i;
+	addrxlat_addr_t nelem;
+	addrxlat_addr_t tblmask;
+	addrxlat_step_t mystep;
+	addrxlat_status status;
+
+	nelem = pf_table_size(&step->meth->param.pgt.pf, step->remain - 1);
+	tblmask = pf_table_mask(&step->meth->param.pgt.pf, step->remain - 1);
+	memcpy(&mystep, step, sizeof *step);
+	while (*addr <= limit) {
+		status = internal_step(step);
+		if (status == ADDRXLAT_ERR_NOTPRESENT) {
+			clear_error(step->ctx);
+			return ADDRXLAT_OK;
+		} else if (status != ADDRXLAT_OK)
+			return status;
+
+		if (step->remain > 1) {
+			status = lowest_unmapped_tbl(step, addr, limit);
+			if (status != ADDRXLAT_ERR_NOTPRESENT)
+				return status;
+		} else
+			*addr = (*addr | tblmask) + 1;
+
+		for (i = 0; i < mystep.remain - 1; ++i)
+			mystep.idx[i] = 0;
+		if (++mystep.idx[i] >= nelem) {
+			bury_cache_buffer(&step->ctx->cache, &mystep.base);
+			break;
+		}
+		memcpy(step, &mystep, sizeof *step);
+	}
+
+	return ADDRXLAT_ERR_NOTPRESENT;
+}
+
+/** Find the lowest unmapped virtual address in a given range.
+ * @param step   Initial step state.
+ * @param addr   First address to try; updated on return.
+ * @param limit  Last address to try.
+ * @returns      Error status.
+ *
+ * The initial step state must be initialized same way as for a call
+ * to @ref addrxlat_launch.
+ */
+addrxlat_status
+lowest_unmapped(addrxlat_step_t *step,
+		addrxlat_addr_t *addr, addrxlat_addr_t limit)
+{
+	addrxlat_addr_t page_mask;
+	addrxlat_status status;
+
+	page_mask = pf_page_mask(&step->meth->param.pgt.pf);
+	*addr &= ~page_mask;
+
+	status = internal_launch(step, *addr);
+	if (status != ADDRXLAT_OK)
+		return status;
+
+	return lowest_unmapped_tbl(step, addr, limit);
+}
+
+/** Find the highest linear mapping in a given range.
+ * @param step   Initial step state.
+ * @param addr   First address to try; updated on return.
+ * @param limit  Last address to try.
+ * @param off    Required virtual-to-kernel-physical offset.
+ * @returns      Error status.
+ *
+ * The initial step state must be initialized same way as for a call
+ * to @ref addrxlat_launch.
+ */
+addrxlat_status
+highest_linear(addrxlat_step_t *step,
+	       addrxlat_addr_t *addr, addrxlat_addr_t limit,
+	       addrxlat_addr_t off)
+{
+	addrxlat_addr_t nextaddr = *addr;
+	addrxlat_status status;
+	addrxlat_status ret = ADDRXLAT_ERR_NOTPRESENT;
+
+	while ( (status = lowest_mapped(step, &nextaddr,
+					limit)) == ADDRXLAT_OK) {
+		addrxlat_fulladdr_t faddr;
+
+		faddr.as = ADDRXLAT_KVADDR;
+		faddr.addr = nextaddr;
+		status = internal_fulladdr_conv(&faddr, ADDRXLAT_KPHYSADDR,
+						step->ctx, step->sys);
+		if (status != ADDRXLAT_OK)
+			return status;
+
+		if (faddr.addr - nextaddr != off)
+			break;
+
+		/* Assume that the whole range is linear. */
+		status = lowest_unmapped(step, &nextaddr, limit);
+		if (status != ADDRXLAT_OK &&
+		    status != ADDRXLAT_ERR_NOTPRESENT)
+			return status;
+		*addr = nextaddr - 1;
+		ret = ADDRXLAT_OK;
+	}
+
+	return (status == ADDRXLAT_OK || status == ADDRXLAT_ERR_NOTPRESENT)
+		? ret
+		: status;
+}
