@@ -42,9 +42,12 @@
 #define _PAGE_PRESENT	(1UL << _PAGE_BIT_PRESENT)
 #define _PAGE_PSE	(1UL << _PAGE_BIT_PSE)
 
-/* Maximum virtual address bits (architecture limit) */
+/** Maximum virtual address bits (architecture limit). */
 #define VIRTADDR_BITS_MAX	48
 
+/** Maximum virtual address bits for 5-level paging (architecture limit). */
+#define VIRTADDR_5L_BITS_MAX	57
+
 /** Page shift (log2 4K). */
 #define PAGE_SHIFT		12
 
@@ -67,114 +70,53 @@
 #define NONCANONICAL_END	(~NONCANONICAL_START)
 #define VIRTADDR_MAX		UINT64_MAX
 
+/* 5-level paging non-canonical extents. */
+#define NONCANONICAL_5L_START	((uint64_t)1<<(VIRTADDR_5L_BITS_MAX-1))
+#define NONCANONICAL_5L_END	(~NONCANONICAL_5L_START)
+
 /** Virtual address of the Xen machine-to-physical map. */
-#define XEN_MACH2PHYS_ADDR	0xffff800000000000ULL
+#define XEN_MACH2PHYS_ADDR	0xffff800000000000
 
-/** Kernel text mapping (virtual addresses).
- * Note that this mapping has never changed, so these constants
- * apply to all kernel versions.
+/** Kernel text mapping (virtual address).
+ * Note that the start address of this mapping has never changed, so this
+ * constant applies to all kernel versions.
  */
-#define __START_KERNEL_map	0xffffffff80000000ULL
-#define __END_KERNEL_map	0xffffffff827fffffULL
+#define LINUX_KTEXT_START	0xffffffff80000000
 
-/* Original Linux layout (before 2.6.11) */
-static const struct sys_region linux_layout_2_6_0[] = {
-	{  0x0000000000000000,  0x0000007fffffffff, /* user space       */
-	   ADDRXLAT_SYS_METH_PGT },
-	/* 0x0000008000000000 - 0x000000ffffffffff     guard hole       */
-	{  0x0000010000000000,  0x000001ffffffffff, /* direct mapping   */
-	   ADDRXLAT_SYS_METH_DIRECT, SYS_ACT_DIRECT },
-	/* 0x0000020000000000 - 0x00007fffffffffff     unused hole      */
-	/* 0x0000800000000000 - 0xffff7fffffffffff     non-canonical    */
-	/* 0xffff800000000000 - 0xfffffeffffffffff     unused hole      */
-	{  0xffffff0000000000,  0xffffff7fffffffff, /* vmalloc/ioremap  */
-	   ADDRXLAT_SYS_METH_PGT },
-	/* 0xffffff8000000000 - 0xffffffff7fffffff     unused hole      */
-	/* 0xffffffff80000000 - 0xffffffff827fffff     kernel text      */
-	/* 0xffffffff82800000 - 0xffffffff9fffffff     unused hole      */
-	{  0xffffffffa0000000,  0xffffffffafffffff, /* modules          */
-	   ADDRXLAT_SYS_METH_PGT },
-	/* 0xffffffffb0000000 - 0xffffffffff5exxxx     unused hole      */
-	{  0xffffffffff5ed000,  0xffffffffffdfffff, /* fixmap/vsyscalls */
-	   ADDRXLAT_SYS_METH_PGT },
-	/* 0xffffffffffe00000 - 0xffffffffffffffff     guard hole       */
-	SYS_REGION_END
-};
+/** Maximum kernel text mapping if kASLR is not active. */
+#define LINUX_KTEXT_END_NOKASLR	0xffffffff9fffffff
 
-/* Linux layout introduced in 2.6.11 */
-static const struct sys_region linux_layout_2_6_11[] = {
-	{  0x0000000000000000,  0x00007fffffffffff, /* user space       */
-	   ADDRXLAT_SYS_METH_PGT },
-	/* 0x0000800000000000 - 0xffff7fffffffffff     non-canonical    */
-	/* 0xffff800000000000 - 0xffff80ffffffffff     guard hole       */
-	{  0xffff810000000000,  0xffffc0ffffffffff, /* direct mapping   */
-	   ADDRXLAT_SYS_METH_DIRECT, SYS_ACT_DIRECT },
-	/* 0xffffc10000000000 - 0xffffc1ffffffffff     guard hole       */
-	{  0xffffc20000000000,  0xffffe1ffffffffff, /* vmalloc/ioremap  */
-	   ADDRXLAT_SYS_METH_PGT },
-	{  0xffffe20000000000,  0xffffe2ffffffffff, /* VMEMMAP          */
-	   ADDRXLAT_SYS_METH_PGT },		    /*   (2.6.24+ only) */
-	/* 0xffffe30000000000 - 0xffffffff7fffffff     unused hole      */
-	/* 0xffffffff80000000 - 0xffffffff827fffff     kernel text      */
-	/* 0xffffffff82800000 - 0xffffffff87ffffff     unused hole      */
-	{  0xffffffff88000000,  0xffffffffffdfffff, /* modules and      */
-	   ADDRXLAT_SYS_METH_PGT },		    /*  fixmap/vsyscall */
-	/* 0xffffffffffe00000 - 0xffffffffffffffff     guard hole       */
-	SYS_REGION_END
-};
+/** Maximum end of kernel text mapping (virtual address).
+ * The kernel text may be smaller, but it must never span beyond this
+ * address.
+ */
+#define LINUX_KTEXT_END		0xffffffffbfffffff
 
-/** Linux layout with hypervisor area, introduced in 2.6.27 */
-static const struct sys_region linux_layout_2_6_27[] = {
-	{  0x0000000000000000,  0x00007fffffffffff, /* user space       */
-	   ADDRXLAT_SYS_METH_PGT },
-	/* 0x0000800000000000 - 0xffff7fffffffffff     non-canonical    */
-	/* 0xffff800000000000 - 0xffff80ffffffffff     guard hole       */
-	/* 0xffff810000000000 - 0xffff87ffffffffff     hypervisor area  */
-	{  0xffff880000000000,  0xffffc0ffffffffff, /* direct mapping   */
-	   ADDRXLAT_SYS_METH_DIRECT, SYS_ACT_DIRECT },
-	/* 0xffffc10000000000 - 0xffffc1ffffffffff     guard hole       */
-	{  0xffffc20000000000,  0xffffe1ffffffffff, /* vmalloc/ioremap  */
-	   ADDRXLAT_SYS_METH_PGT },
-	{  0xffffe20000000000,  0xffffe2ffffffffff, /* VMEMMAP          */
-	   ADDRXLAT_SYS_METH_PGT },
-	/* 0xffffe30000000000 - 0xffffffff7fffffff     unused hole      */
-	/* 0xffffffff80000000 - 0xffffffff827fffff     kernel text      */
-	/* 0xffffffff82800000 - 0xffffffff87ffffff     unused hole      */
-	{  0xffffffff88000000,  0xffffffffffdfffff, /* modules and      */
-	   ADDRXLAT_SYS_METH_PGT },		    /*  fixmap/vsyscall */
-	/* 0xffffffffffe00000 - 0xffffffffffffffff     guard hole       */
-	SYS_REGION_END
-};
+/** Start of direct physical mapping in Linux before 2.6.11 */
+#define LINUX_DIRECTMAP_START_2_6_0	0x0000010000000000
+/** End of direct physical mapping in Linux before 2.6.11 */
+#define LINUX_DIRECTMAP_END_2_6_0	0x000001ffffffffff
 
-/** Linux layout with 64T direct mapping, introduced in 2.6.31 */
-static const struct sys_region linux_layout_2_6_31[] = {
-	{  0x0000000000000000,  0x00007fffffffffff, /* user space       */
-	   ADDRXLAT_SYS_METH_PGT },
-	/* 0x0000800000000000 - 0xffff7fffffffffff     non-canonical    */
-	/* 0xffff800000000000 - 0xffff80ffffffffff     guard hole       */
-	/* 0xffff810000000000 - 0xffff87ffffffffff     hypervisor area  */
-	{  0xffff880000000000,  0xffffc7ffffffffff, /* direct mapping   */
-	   ADDRXLAT_SYS_METH_DIRECT, SYS_ACT_DIRECT },
-	/* 0xffffc80000000000 - 0xffffc8ffffffffff     guard hole       */
-	{  0xffffc90000000000,  0xffffe8ffffffffff, /* vmalloc/ioremap  */
-	   ADDRXLAT_SYS_METH_PGT },
-	/* 0xffffe90000000000 - 0xffffe9ffffffffff     guard hole       */
-	{  0xffffea0000000000,  0xffffeaffffffffff, /* VMEMMAP          */
-	   ADDRXLAT_SYS_METH_PGT },
-	/* 0xffffeb0000000000 - 0xffffffeeffffffff     unused hole      */
-	{  0xffffff0000000000,  0xffffff7fffffffff, /* %esp fixup stack */
-	   ADDRXLAT_SYS_METH_PGT },
-	/* 0xffffff8000000000 - 0xffffffeeffffffff     unused hole      */
-	{  0xffffffef00000000,  0xfffffffeffffffff, /* EFI runtime      */
-	   ADDRXLAT_SYS_METH_PGT },		    /*     (3.14+ only) */
-	/* 0xffffffff00000000 - 0xffffffff7fffffff     guard hole       */
-	/* 0xffffffff80000000 - 0xffffffff827fffff     kernel text      */
-	/* 0xffffffff82800000 - 0xffffffff87ffffff     unused hole      */
-	{  0xffffffff88000000,  0xffffffffffdfffff, /* modules and      */
-	   ADDRXLAT_SYS_METH_PGT },		    /*  fixmap/vsyscall */
-	/* 0xffffffffffe00000 - 0xffffffffffffffff     guard hole       */
-	SYS_REGION_END
-};
+/** Start of direct physical mapping in Linux between 2.6.11 and 2.6.27 */
+#define LINUX_DIRECTMAP_START_2_6_11	0xffff810000000000
+/** End of direct physical mapping in Linux between 2.6.11 and 2.6.27 */
+#define LINUX_DIRECTMAP_END_2_6_11	0xffffc0ffffffffff
+
+/** Start of direct physical mapping in Linux 2.6.27+ */
+#define LINUX_DIRECTMAP_START_2_6_27	0xffff880000000000
+/** End of direct physical mapping in Linux between 2.6.27 and 2.6.31 */
+#define LINUX_DIRECTMAP_END_2_6_27	0xffffc0ffffffffff
+
+/** Start of direct physical mapping in Linux 2.6.31+ */
+#define LINUX_DIRECTMAP_START_2_6_31	LINUX_DIRECTMAP_START_2_6_27
+/** End of direct physical mapping in Linux 2.6.31+ */
+#define LINUX_DIRECTMAP_END_2_6_31	0xffffc7ffffffffff
+
+/** Start of direct physical mapping with 5-level paging */
+#define LINUX_DIRECTMAP_START_5LEVEL	0xff11000000000000
+
+/** End of direct physical mapping with 5-level paging */
+#define LINUX_DIRECTMAP_END_5LEVEL	0xff90ffffffffffff
 
 /** AMD64 (Intel 64) page table step function.
  * @param step  Current step state.
@@ -188,36 +130,41 @@ pgt_x86_64(addrxlat_step_t *step)
 		"Page table",
 		"Page directory",
 		"PDPT table",
+		"PML4 table",
 	};
 	static const char pte_name[][4] = {
 		"pte",
 		"pmd",
 		"pud",
+		"p4d",
 		"pgd",
 	};
+	addrxlat_pte_t pte;
 	addrxlat_status status;
 
-	status = read_pte(step);
+	status = read_pte64(step, &pte);
 	if (status != ADDRXLAT_OK)
 		return status;
 
-	if (!(step->raw.pte & _PAGE_PRESENT))
-		return set_error(step->ctx, ADDRXLAT_ERR_NOTPRESENT,
-				 "%s not present: %s[%u] = 0x%" ADDRXLAT_PRIxPTE,
-				 pgt_full_name[step->remain - 1],
-				 pte_name[step->remain - 1],
-				 (unsigned) step->idx[step->remain],
-				 step->raw.pte);
+	if (!(pte & _PAGE_PRESENT))
+		return !step->ctx->noerr.notpresent
+			? set_error(step->ctx, ADDRXLAT_ERR_NOTPRESENT,
+				    "%s not present: %s[%u] = 0x%" ADDRXLAT_PRIxPTE,
+				    pgt_full_name[step->remain - 1],
+				    pte_name[step->remain - 1],
+				    (unsigned) step->idx[step->remain],
+				    step->raw.pte)
+			: ADDRXLAT_ERR_NOTPRESENT;
 
-	step->base.addr = step->raw.pte & PHYSADDR_MASK;
+	step->base.addr = pte & PHYSADDR_MASK;
 	step->base.as = step->meth->target_as;
 
-	if (step->remain == 3 && (step->raw.pte & _PAGE_PSE)) {
+	if (step->remain == 3 && (pte & _PAGE_PSE)) {
 		step->base.addr &= ~PAGE_MASK_1G;
 		return pgt_huge_page(step);
 	}
 
-	if (step->remain == 2 && (step->raw.pte & _PAGE_PSE)) {
+	if (step->remain == 2 && (pte & _PAGE_PSE)) {
 		step->base.addr &= ~PAGE_MASK_2M;
 		return pgt_huge_page(step);
 	}
@@ -229,34 +176,14 @@ pgt_x86_64(addrxlat_step_t *step)
 	return ADDRXLAT_OK;
 }
 
-/** Get Linux virtual memory layout by kernel version.
- * @param ver  Version code.
- * @returns    Layout definition, or @c NULL.
- */
-static const struct sys_region *
-linux_layout_by_ver(unsigned version_code)
-{
-#define LINUX_LAYOUT_BY_VER(a, b, c)			\
-	if (version_code >= ADDRXLAT_VER_LINUX(a, b, c))	\
-		return linux_layout_ ## a ## _ ## b ## _ ## c
-
-	LINUX_LAYOUT_BY_VER(2, 6, 31);
-	LINUX_LAYOUT_BY_VER(2, 6, 27);
-	LINUX_LAYOUT_BY_VER(2, 6, 11);
-	LINUX_LAYOUT_BY_VER(2, 6, 0);
-
-	return NULL;
-}
-
-/** Check whether a virtual address is mapped to a physical address.
+/** Translate virtual to kernel physical using page tables.
  * @param sys    Translation system object.
- * @param ctx    Address translation context.
- * @param addr   Address to be checked.
- * @returns      Non-zero if the address can be translated.
+ * @param ctx    Address translation object.
+ * @param addr   Address to be translated.
+ * @returns      Error status.
  */
-static int
-is_mapped(addrxlat_sys_t *sys, addrxlat_ctx_t *ctx,
-	  addrxlat_addr_t addr)
+static addrxlat_status
+vtop_pgt(addrxlat_sys_t *sys, addrxlat_ctx_t *ctx, addrxlat_addr_t *addr)
 {
 	addrxlat_step_t step;
 	addrxlat_status status;
@@ -264,13 +191,64 @@ is_mapped(addrxlat_sys_t *sys, addrxlat_ctx_t *ctx,
 	step.ctx = ctx;
 	step.sys = sys;
 	step.meth = &sys->meth[ADDRXLAT_SYS_METH_PGT];
-	status = internal_launch(&step, addr);
+	step.base.addr = *addr;
+	status = internal_walk(&step);
+	if (status != ADDRXLAT_OK)
+		return status;
 
+	status = internal_fulladdr_conv(&step.base, ADDRXLAT_KPHYSADDR,
+					ctx, sys);
 	if (status == ADDRXLAT_OK)
-		status = internal_walk(&step);
+		*addr = step.base.addr;
 
-	clear_error(ctx);
-	return status == ADDRXLAT_OK;
+	return status;
+}
+
+/** Remove the method for the reverse direct mapping.
+ * @param sys  Translation system.
+ */
+static void
+remove_rdirect(addrxlat_sys_t *sys)
+{
+	addrxlat_map_t *map;
+
+	sys->meth[ADDRXLAT_SYS_METH_RDIRECT].kind = ADDRXLAT_NOMETH;
+	map = sys->map[ADDRXLAT_SYS_MAP_KPHYS_DIRECT];
+	if (map) {
+		internal_map_decref(map);
+		sys->map[ADDRXLAT_SYS_MAP_KPHYS_DIRECT] = NULL;
+	}
+}
+
+/** Get Linux directmap layout by kernel version.
+ * @param rgn  Directmap region; updated on success.
+ * @param ver  Version code.
+ * @returns    Error status.
+ *
+ * The @c first and @c last fields of @c rgn are set according to
+ * the Linux kernel version if this function returns @ref ADDRXLAT_OK.
+ * No error message is set if this function fails, so the caller need
+ * not clear it with @ref clear_error.
+ */
+static addrxlat_status
+linux_directmap_by_ver(struct sys_region *rgn, unsigned ver)
+{
+	if (ver >= ADDRXLAT_VER_LINUX(4, 8, 0))
+		return ADDRXLAT_ERR_NOMETH;
+
+#define LINUX_DIRECTMAP_BY_VER(a, b, c)			\
+	if (ver >= ADDRXLAT_VER_LINUX(a, b, c)) {		\
+		rgn->first = LINUX_DIRECTMAP_START_ ##a##_##b##_##c;	\
+		rgn->last = LINUX_DIRECTMAP_END_ ##a##_##b##_##c;	\
+		return ADDRXLAT_OK;					\
+	}
+
+	LINUX_DIRECTMAP_BY_VER(2, 6, 31);
+	LINUX_DIRECTMAP_BY_VER(2, 6, 27);
+	LINUX_DIRECTMAP_BY_VER(2, 6, 11);
+	LINUX_DIRECTMAP_BY_VER(2, 6, 0);
+
+	return ADDRXLAT_ERR_NOTIMPL;
 }
 
 /** Check whether an address looks like start of direct mapping.
@@ -283,84 +261,117 @@ static int
 is_directmap(addrxlat_sys_t *sys, addrxlat_ctx_t *ctx,
 	     addrxlat_addr_t addr)
 {
-	addrxlat_fulladdr_t faddr;
+	addrxlat_status status = vtop_pgt(sys, ctx, &addr);
+	clear_error(ctx);
+	return status == ADDRXLAT_OK && addr == 0;
+}
+
+/** Get directmap location by walking page tables.
+ * @param rgn  Directmap region; updated on success.
+ * @param sys  Translation system object.
+ * @param ctx  Address translation context.
+ * @returns    Error status.
+ *
+ * No error message is set if this function fails, so the caller need
+ * not clear it with @ref clear_error.
+ */
+static addrxlat_status
+linux_directmap_by_pgt(struct sys_region *rgn,
+		       addrxlat_sys_t *sys, addrxlat_ctx_t *ctx)
+{
+	addrxlat_step_t step;
+	addrxlat_addr_t end;
 	addrxlat_status status;
 
-	faddr.addr = addr;
-	faddr.as = ADDRXLAT_KVADDR;
-	status = internal_fulladdr_conv(&faddr, ADDRXLAT_KPHYSADDR, ctx, sys);
-	clear_error(ctx);
-	return status == ADDRXLAT_OK && faddr.addr == 0;
+	step.ctx = ctx;
+	step.sys = sys;
+	step.meth = &sys->meth[ADDRXLAT_SYS_METH_PGT];
+
+	if (is_directmap(sys, ctx, LINUX_DIRECTMAP_START_2_6_0)) {
+		rgn->last = rgn->first = LINUX_DIRECTMAP_START_2_6_0;
+		return highest_linear(&step, &rgn->last,
+				      LINUX_DIRECTMAP_END_2_6_0, -rgn->first);
+	}
+
+	if (is_directmap(sys, ctx, LINUX_DIRECTMAP_START_2_6_11)) {
+		rgn->last = rgn->first = LINUX_DIRECTMAP_START_2_6_11;
+		return highest_linear(&step, &rgn->last,
+				      LINUX_DIRECTMAP_END_2_6_11, -rgn->first);
+	}
+
+	if (step.meth->param.pgt.pf.nfields == 6) {
+		rgn->first = LINUX_DIRECTMAP_START_5LEVEL;
+		end = LINUX_DIRECTMAP_END_5LEVEL;
+	} else {
+		rgn->first = LINUX_DIRECTMAP_START_2_6_31;
+		end = LINUX_DIRECTMAP_END_2_6_31;
+	}
+	status = lowest_mapped(&step, &rgn->first, end);
+	if (status == ADDRXLAT_OK) {
+		rgn->last = rgn->first;
+		return highest_linear(&step, &rgn->last, end, -rgn->first);
+	}
+
+	return ADDRXLAT_ERR_NOTIMPL;
 }
 
-/** Get virtual memory layout by walking page tables.
- * @param sys    Translation system object.
- * @param ctx    Address translation context.
- * @returns      Memory layout, or @c NULL if undetermined.
+/** Set up Linux direct mapping on x86_64.
+ * @param ctl  Initialization data.
+ * @returns    Error status.
  */
-static const struct sys_region *
-linux_layout_by_pgt(addrxlat_sys_t *sys, addrxlat_ctx_t *ctx)
+static addrxlat_status
+linux_directmap(struct os_init_data *ctl)
 {
-	/* Only pre-2.6.11 kernels had this direct mapping */
-	if (is_directmap(sys, ctx, 0x0000010000000000))
-		return linux_layout_2_6_0;
+	struct sys_region layout[2];
+	addrxlat_status status;
 
-	/* Only kernels between 2.6.11 and 2.6.27 had this direct mapping */
-	if (is_directmap(sys, ctx, 0xffff810000000000))
-		return linux_layout_2_6_11;
-
-	/* Only 2.6.31+ kernels map VMEMMAP at this address */
-	if (is_mapped(sys, ctx, 0xffffea0000000000))
-		return linux_layout_2_6_31;
-
-	/* Sanity check for 2.6.27+ direct mapping */
-	if (is_directmap(sys, ctx, 0xffff880000000000))
-		return linux_layout_2_6_27;
-
-	return NULL;
+	status = linux_directmap_by_pgt(&layout[0], ctl->sys, ctl->ctx);
+	if (status != ADDRXLAT_OK && ctl->osdesc->ver)
+		status = linux_directmap_by_ver(&layout[0], ctl->osdesc->ver);
+	remove_rdirect(ctl->sys);
+	if (status == ADDRXLAT_OK) {
+		layout[0].meth = ADDRXLAT_SYS_METH_DIRECT;
+		layout[0].act = SYS_ACT_DIRECT;
+		layout[1].meth = ADDRXLAT_SYS_METH_NUM;
+		status = sys_set_layout(ctl, ADDRXLAT_SYS_MAP_KV_PHYS, layout);
+		if (status != ADDRXLAT_OK)
+			return status;
+	}
+	return ADDRXLAT_OK;
 }
 
-/** Set Linux kernel text mapping offset.
+/** Set the kernel text mapping offset.
+ * @param sys  Translation system object.
+ * @param off  Offset from physical to virtual addresses.
+ */
+static void
+set_ktext_offset(addrxlat_sys_t *sys, addrxlat_addr_t off)
+{
+	addrxlat_meth_t *meth = &sys->meth[ADDRXLAT_SYS_METH_KTEXT];
+	meth->kind = ADDRXLAT_LINEAR;
+	meth->target_as = ADDRXLAT_KPHYSADDR;
+	meth->param.linear.off = off;
+}
+
+/** Calculate Linux kernel text mapping offset using page tables.
  * @param sys    Translation system object.
  * @param ctx    Address translation object.
  * @param vaddr  Any valid kernel text virtual address.
  * @returns      Error status.
  */
 static addrxlat_status
-set_ktext_offset(addrxlat_sys_t *sys, addrxlat_ctx_t *ctx,
-		 addrxlat_addr_t vaddr)
+calc_ktext_offset(addrxlat_sys_t *sys, addrxlat_ctx_t *ctx,
+		  addrxlat_addr_t vaddr)
 {
-	addrxlat_sys_meth_t methidx;
-	addrxlat_step_t step;
-	addrxlat_meth_t *meth;
+	addrxlat_addr_t paddr;
 	addrxlat_status status;
 
-	methidx = internal_map_search(sys->map[ADDRXLAT_SYS_MAP_HW], vaddr);
-	if (methidx == ADDRXLAT_SYS_METH_NONE)
-		return set_error(ctx, ADDRXLAT_ERR_NOMETH,
-				 "No translation for %"ADDRXLAT_PRIxADDR,
-				 vaddr);
-
-	step.ctx = ctx;
-	step.sys = sys;
-	step.meth = &sys->meth[methidx];
-	status = internal_launch(&step, vaddr);
+	paddr = vaddr;
+	status = vtop_pgt(sys, ctx, &paddr);
 	if (status != ADDRXLAT_OK)
 		return status;
 
-	status = internal_walk(&step);
-	if (status != ADDRXLAT_OK)
-		return status;
-
-	status = internal_fulladdr_conv(&step.base, ADDRXLAT_KPHYSADDR,
-					ctx, sys);
-	if (status != ADDRXLAT_OK)
-		return status;
-
-	meth = &sys->meth[ADDRXLAT_SYS_METH_KTEXT];
-	meth->kind = ADDRXLAT_LINEAR;
-	meth->target_as = ADDRXLAT_KPHYSADDR;
-	meth->param.linear.off = step.base.addr - vaddr;
+	set_ktext_offset(sys, paddr - vaddr);
 	return ADDRXLAT_OK;
 }
 
@@ -379,51 +390,164 @@ set_pgt_fallback(addrxlat_sys_t *sys, addrxlat_sys_meth_t idx)
 		*meth = sys->meth[ADDRXLAT_SYS_METH_PGT];
 }
 
-/** The beginning of the kernel text virtual mapping may not be mapped
- * for various reasons. Let's use an offset of 16M to be safe.
+/** Set up Linux kernel reverse direct mapping on x86_64.
+ * @param ctl  Initialization data.
+ * @returns    Error status.
  */
-#define LINUX_KTEXT_SKIP		(16ULL << 20)
+static addrxlat_status
+linux_rdirect_map(struct os_init_data *ctl)
+{
+	/** Possible direct mapping locations (if not randomized). */
+	static const addrxlat_addr_t fixed_loc[] = {
+		LINUX_DIRECTMAP_START_2_6_0,
+		LINUX_DIRECTMAP_START_2_6_11,
+		LINUX_DIRECTMAP_START_2_6_31,
+	};
 
-/** Xen kernels are loaded low in memory. The ktext mapping may not go up
- * to 16M then. Let's use 1M, because Xen kernel should take up at least
- * 1M of RAM, and this value also covers kernels loaded at 1M (so this code
- * may be potentially reused for ia32).
- */
-#define LINUX_KTEXT_SKIP_alt		(1ULL << 20)
+	struct sys_region layout[2];
+	int i;
+	addrxlat_fulladdr_t page_offset;
+	addrxlat_status status;
+
+	if (!ctl->ctx->cb.get_page ||
+	    !(ctl->ctx->cb.read_caps & ADDRXLAT_CAPS(ADDRXLAT_KVADDR)))
+		return ADDRXLAT_ERR_NOMETH;
+
+	layout[0].first = 0;
+	layout[0].last = PHYSADDR_MASK;
+	layout[0].meth = ADDRXLAT_SYS_METH_RDIRECT;
+	layout[0].act = SYS_ACT_RDIRECT;
+	layout[1].meth = ADDRXLAT_SYS_METH_NUM;
+
+	status = get_symval(ctl->ctx, "page_offset_base",
+			    &page_offset.addr);
+	if (status == ADDRXLAT_OK) {
+		uint64_t val;
+
+		page_offset.as = ADDRXLAT_KVADDR;
+		status = do_read64(ctl->ctx, &page_offset, &val);
+		if (status != ADDRXLAT_OK)
+			return status;
+
+		ctl->sys->meth[ADDRXLAT_SYS_METH_DIRECT]
+			.param.linear.off = -val;
+		status = sys_set_layout(
+			ctl, ADDRXLAT_SYS_MAP_KPHYS_DIRECT, layout);
+		if (status != ADDRXLAT_OK)
+			goto err_layout;
+
+		if (is_directmap(ctl->sys, ctl->ctx, val))
+			return ADDRXLAT_OK;
+	}
+
+	for (i = 0; i < ARRAY_SIZE(fixed_loc); ++i) {
+		ctl->sys->meth[ADDRXLAT_SYS_METH_DIRECT]
+			.param.linear.off = -fixed_loc[i];
+		status = sys_set_layout(
+			ctl, ADDRXLAT_SYS_MAP_KPHYS_DIRECT, layout);
+		if (status != ADDRXLAT_OK)
+			goto err_layout;
+
+		if (is_directmap(ctl->sys, ctl->ctx, fixed_loc[i]))
+			return ADDRXLAT_OK;
+
+		remove_rdirect(ctl->sys);
+	}
+
+	return ADDRXLAT_NOMETH;
+
+ err_layout:
+	return set_error(ctl->ctx, status, "Cannot set up %s",
+			 "Linux kernel direct mapping");
+}
 
 /** Set up Linux kernel text translation method.
- * @param ctl     Initialization data.
- * @param region  Associated region definition.
+ * @param ctl  Initialization data.
+ * @returns    Error status.
  */
 static addrxlat_status
 linux_ktext_meth(struct os_init_data *ctl)
 {
+	addrxlat_addr_t stext;
 	addrxlat_status status;
 
-	if (ctl->popt.val[OPT_physbase].set) {
-		addrxlat_meth_t *meth =
-			&ctl->sys->meth[ADDRXLAT_SYS_METH_KTEXT];
-
-		meth->kind = ADDRXLAT_LINEAR;
-		meth->target_as = ADDRXLAT_KPHYSADDR;
-		meth->param.linear.off = ctl->popt.val[OPT_physbase].num -
-			__START_KERNEL_map;
+	if (ctl->popt.val[OPT_phys_base].set) {
+		set_ktext_offset(ctl->sys, (ctl->popt.val[OPT_phys_base].addr -
+					    LINUX_KTEXT_START));
 		return ADDRXLAT_OK;
 	}
 
-	status = set_ktext_offset(ctl->sys, ctl->ctx,
-				  __START_KERNEL_map + LINUX_KTEXT_SKIP);
-	if (status == ADDRXLAT_ERR_NOTPRESENT || status == ADDRXLAT_ERR_NODATA) {
+	status = get_symval(ctl->ctx, "_stext", &stext);
+	if (status == ADDRXLAT_ERR_NODATA) {
 		clear_error(ctl->ctx);
-		status = set_ktext_offset(ctl->sys, ctl->ctx,
-					  __START_KERNEL_map +
-					  LINUX_KTEXT_SKIP_alt);
+		status = get_symval(ctl->ctx, "_text", &stext);
 	}
+	if (status == ADDRXLAT_ERR_NODATA) {
+		addrxlat_step_t step;
+
+		clear_error(ctl->ctx);
+		step.ctx = ctl->ctx;
+		step.sys = ctl->sys;
+		step.meth = &ctl->sys->meth[ADDRXLAT_SYS_METH_PGT];
+		stext = LINUX_KTEXT_START;
+		status = lowest_mapped(&step, &stext, LINUX_KTEXT_END);
+		if (status != ADDRXLAT_OK)
+			return status;
+
+		status = internal_fulladdr_conv(&step.base, ADDRXLAT_KPHYSADDR,
+						step.ctx, step.sys);
+		if (status != ADDRXLAT_OK)
+			return status;
+
+		set_ktext_offset(ctl->sys, step.base.addr - stext);
+		return ADDRXLAT_OK;
+	} else if (status == ADDRXLAT_OK)
+		status = calc_ktext_offset(ctl->sys, ctl->ctx, stext);
+
 	if (status != ADDRXLAT_OK)
 		return set_error(ctl->ctx, status, "Cannot translate ktext");
 	return status;
 }
 
+/** Find the kernel text mapping extents.
+ * @param ctl   Initialization data.
+ * @param low   Lowest ktext address (set on successful return).
+ * @param high  Highest ktext address (set on successful return).
+ * @returns     Error status.
+ */
+static addrxlat_status
+linux_ktext_extents(struct os_init_data *ctl,
+		    addrxlat_addr_t *low, addrxlat_addr_t *high)
+{
+	addrxlat_addr_t linearoff;
+	addrxlat_step_t step;
+	addrxlat_status status;
+
+	step.ctx = ctl->ctx;
+	step.sys = ctl->sys;
+	step.meth = &ctl->sys->meth[ADDRXLAT_SYS_METH_PGT];
+	*low = LINUX_KTEXT_START;
+	status = lowest_mapped(&step, low, LINUX_KTEXT_END);
+	if (status != ADDRXLAT_OK)
+		return status;
+
+	linearoff = ctl->sys->meth[ADDRXLAT_SYS_METH_KTEXT].param.linear.off;
+	*high = *low;
+	status = highest_linear(&step, high, LINUX_KTEXT_END_NOKASLR,
+				linearoff);
+	if (status == ADDRXLAT_OK && *high == LINUX_KTEXT_END_NOKASLR) {
+		++*high;
+		status = highest_linear(&step, high, LINUX_KTEXT_END,
+					linearoff);
+		if (status == ADDRXLAT_ERR_NOTPRESENT) {
+			clear_error(step.ctx);
+			--*high;
+			status = ADDRXLAT_OK;
+		}
+	}
+	return status;
+}
+
 /** Set up Linux kernel text mapping on x86_64.
  * @param ctl  Initialization data.
  * @returns    Error status.
@@ -432,23 +556,54 @@ static addrxlat_status
 linux_ktext_map(struct os_init_data *ctl)
 {
 	addrxlat_range_t range;
+	addrxlat_meth_t *meth;
+	addrxlat_addr_t low, high;
 	addrxlat_status status;
 
 	status = linux_ktext_meth(ctl);
-	if (status != ADDRXLAT_OK &&
-	    status != ADDRXLAT_ERR_NOMETH &&
-	    status != ADDRXLAT_ERR_NODATA &&
-	    status != ADDRXLAT_ERR_NOTPRESENT)
-		return status;
-	clear_error(ctl->ctx);
+	if (status == ADDRXLAT_ERR_NOMETH ||
+	    status == ADDRXLAT_ERR_NODATA ||
+	    status == ADDRXLAT_ERR_NOTPRESENT)
+		goto err_nonfatal;
+	else if (status != ADDRXLAT_OK)
+		return set_error(ctl->ctx, status, "Cannot set up %s",
+				 "Linux kernel text mapping");
 
-	range.endoff = __END_KERNEL_map - __START_KERNEL_map;
-	range.meth = ADDRXLAT_SYS_METH_KTEXT;
-	status = internal_map_set(ctl->sys->map[ADDRXLAT_SYS_MAP_KV_PHYS],
-				  __START_KERNEL_map, &range);
-	if (status != ADDRXLAT_OK)
+	meth = &ctl->sys->meth[ADDRXLAT_SYS_METH_PGT];
+	if (meth->kind == ADDRXLAT_PGT &&
+	    meth->param.pgt.root.as == ADDRXLAT_KVADDR) {
+		/* minimal ktext mapping for the root page table */
+		range.endoff = PAGE_MASK;
+		range.meth = ADDRXLAT_SYS_METH_KTEXT;
+		status = internal_map_set(
+			ctl->sys->map[ADDRXLAT_SYS_MAP_KV_PHYS],
+			meth->param.pgt.root.addr, &range);
+		if (status != ADDRXLAT_OK)
+			return set_error(ctl->ctx, status, "Cannot set up %s",
+					 "minimal Linux kernel text mapping");
+	}
+
+	status = linux_ktext_extents(ctl, &low, &high);
+	if (status == ADDRXLAT_ERR_NOMETH ||
+	    status == ADDRXLAT_ERR_NODATA ||
+	    status == ADDRXLAT_ERR_NOTPRESENT)
+		goto err_nonfatal;
+	else if (status != ADDRXLAT_OK)
 		return set_error(ctl->ctx, status,
-				 "Cannot set up Linux kernel text mapping");
+				 "Linux kernel text search failed");
+
+	range.meth = ADDRXLAT_SYS_METH_KTEXT;
+	range.endoff = high - low;
+	status = internal_map_set(
+		ctl->sys->map[ADDRXLAT_SYS_MAP_KV_PHYS], low, &range);
+	if (status != ADDRXLAT_OK)
+		return set_error(ctl->ctx, status, "Cannot set up %s",
+				 "Linux kernel text mapping");
+
+	return ADDRXLAT_OK;
+
+ err_nonfatal:
+	clear_error(ctl->ctx);
 	return ADDRXLAT_OK;
 }
 
@@ -501,6 +656,7 @@ set_xen_p2m(struct os_init_data *ctl)
 	meth->target_as = ADDRXLAT_MACHPHYSADDR;
 	meth->param.pgt.root.addr = p2m_maddr;
 	meth->param.pgt.root.as = ADDRXLAT_MACHPHYSADDR;
+	meth->param.pgt.pte_mask = 0;
 	meth->param.pgt.pf = xen_p2m_pf;
 
 	range.endoff = paging_max_index(&xen_p2m_pf);
@@ -520,11 +676,22 @@ set_xen_p2m(struct os_init_data *ctl)
 static addrxlat_status
 map_linux_x86_64(struct os_init_data *ctl)
 {
-	const struct sys_region *layout;
+	static const struct sym_spec pgtspec[] = {
+		{ ADDRXLAT_SYM_REG, ADDRXLAT_MACHPHYSADDR, "cr3" },
+		{ ADDRXLAT_SYM_VALUE, ADDRXLAT_KVADDR, "init_top_pgt" },
+		{ ADDRXLAT_SYM_VALUE, ADDRXLAT_KVADDR, "init_level4_pgt" },
+		{ ADDRXLAT_SYM_NONE }
+	};
+
+	addrxlat_meth_t *meth;
 	addrxlat_status status;
 
-	sys_sym_pgtroot(ctl, "cr3", "init_level4_pgt");
+	/* Set up page table translation. */
+	sys_sym_pgtroot(ctl, pgtspec);
+	meth = &ctl->sys->meth[ADDRXLAT_SYS_METH_PGT];
+	meth->param.pgt.root.addr &= ~PAGE_MASK;
 
+	/* Take care of machine physical <-> kernel physical mapping. */
 	if (ctl->popt.val[OPT_xen_xlat].set &&
 	    ctl->popt.val[OPT_xen_xlat].num) {
 		status = set_xen_p2m(ctl);
@@ -534,23 +701,32 @@ map_linux_x86_64(struct os_init_data *ctl)
 		set_xen_mach2phys(ctl, XEN_MACH2PHYS_ADDR);
 	}
 
+	/* Make sure physical addresses can be accessed.
+	 * This is crucial for page table translation.
+	 */
+	if (!(ctl->ctx->cb.read_caps & ADDRXLAT_CAPS(ADDRXLAT_MACHPHYSADDR)) &&
+	    !(ctl->ctx->cb.read_caps & ADDRXLAT_CAPS(ADDRXLAT_KPHYSADDR))) {
+		status = linux_rdirect_map(ctl);
+		if (status != ADDRXLAT_OK &&
+		    status != ADDRXLAT_ERR_NOMETH &&
+		    status != ADDRXLAT_ERR_NODATA &&
+		    status != ADDRXLAT_ERR_NOTPRESENT)
+			return status;
+
+		clear_error(ctl->ctx);
+	}
+
+	/* Set up kernel text mapping. */
 	status = linux_ktext_map(ctl);
 	if (status != ADDRXLAT_OK)
 		return status;
+	set_pgt_fallback(ctl->sys, ADDRXLAT_SYS_METH_KTEXT);
 
-	layout = linux_layout_by_pgt(ctl->sys, ctl->ctx);
-
-	if (!layout && ctl->osdesc->ver)
-		layout = linux_layout_by_ver(ctl->osdesc->ver);
-	if (!layout)
-		return ADDRXLAT_OK;
-
-	status = sys_set_layout(ctl, ADDRXLAT_SYS_MAP_KV_PHYS, layout);
+	/* Set up direct mapping. */
+	status = linux_directmap(ctl);
 	if (status != ADDRXLAT_OK)
 		return status;
 
-	set_pgt_fallback(ctl->sys, ADDRXLAT_SYS_METH_KTEXT);
-
 	return ADDRXLAT_OK;
 }
 
@@ -620,23 +796,30 @@ is_xen_ktext(struct os_init_data *ctl, addrxlat_addr_t addr)
 static addrxlat_status
 setup_xen_pgt(struct os_init_data *ctl)
 {
+	static const struct sym_spec pgtspec[] = {
+		{ ADDRXLAT_SYM_REG, ADDRXLAT_MACHPHYSADDR, "cr3" },
+		{ ADDRXLAT_SYM_VALUE, ADDRXLAT_KVADDR, "pgd_l4" },
+		{ ADDRXLAT_SYM_NONE }
+	};
+
 	struct sys_region layout[2];
 	addrxlat_meth_t *meth;
 	addrxlat_addr_t pgt;
 	addrxlat_off_t off;
 	addrxlat_status status;
 
-	status = sys_sym_pgtroot(ctl, "cr3", "pgd_l4");
+	status = sys_sym_pgtroot(ctl, pgtspec);
 	meth = &ctl->sys->meth[ADDRXLAT_SYS_METH_PGT];
+	meth->param.pgt.root.addr &= ~PAGE_MASK;
 	if (meth->param.pgt.root.as != ADDRXLAT_KVADDR)
 		return status;	/* either unset or physical */
 
 	pgt = meth->param.pgt.root.addr;
 	if (pgt >= XEN_DIRECTMAP) {
 		off = -XEN_DIRECTMAP;
-	} else if (ctl->popt.val[OPT_physbase].set) {
+	} else if (ctl->popt.val[OPT_phys_base].set) {
 		addrxlat_addr_t xen_virt_start = pgt & ~(XEN_TEXT_SIZE - 1);
-		off = ctl->popt.val[OPT_physbase].num - xen_virt_start;
+		off = ctl->popt.val[OPT_phys_base].addr - xen_virt_start;
 	} else
 		return ADDRXLAT_ERR_NODATA;
 
@@ -652,10 +835,7 @@ setup_xen_pgt(struct os_init_data *ctl)
 	if (status != ADDRXLAT_OK)
 		return status;
 
-	meth = &ctl->sys->meth[ADDRXLAT_SYS_METH_KTEXT];
-	meth->kind = ADDRXLAT_LINEAR;
-	meth->target_as = ADDRXLAT_KPHYSADDR;
-	meth->param.linear.off = off;
+	set_ktext_offset(ctl->sys, off);
 	return ADDRXLAT_OK;
 }
 
@@ -732,7 +912,7 @@ map_xen_x86_64(struct os_init_data *ctl)
 		return status;
 
 	if (layout[1].meth == ADDRXLAT_SYS_METH_KTEXT) {
-		set_ktext_offset(ctl->sys, ctl->ctx, layout[1].first);
+		calc_ktext_offset(ctl->sys, ctl->ctx, layout[1].first);
 		clear_error(ctl->ctx);
 		set_pgt_fallback(ctl->sys, ADDRXLAT_SYS_METH_KTEXT);
 	}
@@ -750,6 +930,16 @@ static const struct sys_region layout_generic[] = {
 	SYS_REGION_END
 };
 
+/** Generic x86_64 5-level paging layout */
+static const struct sys_region layout_5level[] = {
+	{  0,  NONCANONICAL_5L_START - 1,		/* lower half       */
+	   ADDRXLAT_SYS_METH_PGT },
+	/* NONCANONICAL_5L_START - NONCANONICAL_5L_END   non-canonical    */
+	{  NONCANONICAL_5L_END + 1,  VIRTADDR_MAX,	/* higher half      */
+	   ADDRXLAT_SYS_METH_PGT },
+	SYS_REGION_END
+};
+
 /** Initialize a translation map for an x86_64 OS.
  * @param ctl  Initialization data.
  * @returns    Error status.
@@ -760,7 +950,7 @@ sys_x86_64(struct os_init_data *ctl)
 	static const addrxlat_paging_form_t x86_64_pf = {
 		.pte_format = ADDRXLAT_PTE_X86_64,
 		.nfields = 5,
-		.fieldsz = { 12, 9, 9, 9, 9 }
+		.fieldsz = { 12, 9, 9, 9, 9, 9 }
 	};
 	addrxlat_map_t *map;
 	addrxlat_meth_t *meth;
@@ -773,9 +963,21 @@ sys_x86_64(struct os_init_data *ctl)
 		meth->param.pgt.root = ctl->popt.val[OPT_rootpgt].fulladdr;
 	else
 		meth->param.pgt.root.as = ADDRXLAT_NOADDR;
+	meth->param.pgt.pte_mask =
+		opt_num_default(&ctl->popt, OPT_pte_mask, 0);
 	meth->param.pgt.pf = x86_64_pf;
 
-	status = sys_set_layout(ctl, ADDRXLAT_SYS_MAP_HW, layout_generic);
+	if (ctl->popt.val[OPT_levels].set) {
+		long levels = ctl->popt.val[OPT_levels].num;
+		if (levels < 4 || levels > 5)
+			return bad_paging_levels(ctl->ctx, levels);
+		meth->param.pgt.pf.nfields = levels + 1;
+	}
+
+	status = sys_set_layout(ctl, ADDRXLAT_SYS_MAP_HW,
+				(meth->param.pgt.pf.nfields == 6
+				 ? layout_5level
+				 : layout_generic));
 	if (status != ADDRXLAT_OK)
 		return status;
 
