@@ -34,7 +34,7 @@
 #include <limits.h>
 #include <ctype.h>
 #include <unistd.h>
-#include <addrxlat.h>
+#include <libkdumpfile/addrxlat.h>
 
 #include "testutil.h"
 
@@ -142,6 +142,7 @@ enum param_index {
 
 	/* Page tables */
 	pgt_root,
+	pgt_pte_mask,
 	pgt_pte_format,
 	pgt_fields,
 
@@ -169,6 +170,7 @@ static const kw_pair_t linear_param[] = {
 
 static const kw_pair_t pgt_param[] = {
 	{"root", pgt_root },
+	{"pte_mask", pgt_pte_mask },
 	{"pte_format", pgt_pte_format },
 	{"fields", pgt_fields },
 	{NULL}
@@ -464,6 +466,14 @@ parse_meth_param(const char *spec, addrxlat_meth_t *meth)
 		res = parse_fulladdr(p, &meth->param.pgt.root);
 		break;
 
+	case pgt_pte_mask:
+		meth->param.pgt.pte_mask = strtoull(p, (char**)&endp, 0);
+		if (endp != p && !*endp)
+			res = TEST_OK;
+		else
+			fprintf(stderr, "Invalid number: %s", p);
+		break;
+
 	case pgt_pte_format:
 		res = parse_pte_format(p, &meth->param.pgt.pf.pte_format);
 		break;
@@ -750,13 +760,6 @@ struct entry {
 	char buf[];
 };
 
-enum read_status {
-	read_ok = ADDRXLAT_OK,
-	read_notfound,
-	read_vtop_failed,
-	read_unknown_as,
-};
-
 struct cbdata {
 	addrxlat_ctx_t *ctx;
 	addrxlat_sys_t *sys;
@@ -775,42 +778,25 @@ find_entry(addrxlat_addr_t addr, size_t sz)
 }
 
 static addrxlat_status
-read32(void *data, const addrxlat_fulladdr_t *addr, uint32_t *val)
-{
-	struct cbdata *cbd = data;
-	struct entry *ent;
-	uint32_t *p;
-
-	if (addr->as != ADDRXLAT_MACHPHYSADDR)
-		return addrxlat_ctx_err(cbd->ctx, ADDRXLAT_ERR_INVALID,
-					"Unexpected address space: %ld",
-					(long)addr->as);
-
-	ent = find_entry(addr->addr, sizeof(uint32_t));
-	if (!ent)
-		return addrxlat_ctx_err(cbd->ctx, -read_notfound, "No data");
-	p = (uint32_t*)(ent->buf + addr->addr - ent->addr);
-	*val = *p;
-	return ADDRXLAT_OK;
-}
-
-static addrxlat_status
-read64(void *data, const addrxlat_fulladdr_t *addr, uint64_t *val)
+get_page(void *data, addrxlat_buffer_t *buf)
 {
 	struct cbdata *cbd = data;
 	struct entry *ent;
 	uint64_t *p;
 
-	if (addr->as != ADDRXLAT_MACHPHYSADDR)
+	if (buf->addr.as != ADDRXLAT_MACHPHYSADDR)
 		return addrxlat_ctx_err(cbd->ctx, ADDRXLAT_ERR_INVALID,
 					"Unexpected address space: %ld",
-					(long)addr->as);
+					(long)buf->addr.as);
 
-	ent = find_entry(addr->addr, sizeof(uint64_t));
+	ent = find_entry(buf->addr.addr, sizeof(uint64_t));
 	if (!ent)
-		return addrxlat_ctx_err(cbd->ctx, -read_notfound, "No data");
-	p = (uint64_t*)(ent->buf + addr->addr - ent->addr);
-	*val = *p;
+		return addrxlat_ctx_err(cbd->ctx, ADDRXLAT_ERR_NODATA,
+					"No data");
+	buf->addr.addr = ent->addr;
+	buf->ptr = ent->buf;
+	buf->size = ent->buflen;
+	buf->byte_order = ADDRXLAT_HOST_ENDIAN;
 	return ADDRXLAT_OK;
 }
 
@@ -920,8 +906,7 @@ int main(int argc, char *argv[])
 	struct cbdata data;
 	addrxlat_cb_t cb = {
 		.data = &data,
-		.read32 = read32,
-		.read64 = read64,
+		.get_page = get_page,
 		.read_caps = ADDRXLAT_CAPS(ADDRXLAT_MACHPHYSADDR)
 	};
 	int c;
@@ -975,6 +960,8 @@ int main(int argc, char *argv[])
 	}
 	data.sys = read_config(cfg);
 	fclose(cfg);
+	if (!data.sys)
+		return TEST_ERR;
 
 	for (i = optind; i < argc; ++i) {
 		rc = translate(&data, argv[i]);
